@@ -26,11 +26,11 @@ finds a match it moves onto handshake2
 
 handshake2: client sends sha25(remote_seed + auth_hash) to
 the device along with the TP_SESSIONID.  Device responds with
-200 if succesful.  It generally will be because this
-implemenation checks the auth_hash it recevied during handshake1
+200 if successful.  It generally will be because this
+implementation checks the auth_hash it received during handshake1
 
 encryption: local_seed, remote_seed and auth_hash are now used
-for encryption.  The last 4 bytes of the initialisation vector
+for encryption.  The last 4 bytes of the initialization vector
 are used as a sequence number that increments every time the
 client calls encrypt and this sequence number is sent as a
 url parameter to the device along with the encrypted payload
@@ -40,40 +40,45 @@ https://github.com/python-kasa/python-kasa/pull/117
 
 """
 
+from __future__ import annotations
+
 import asyncio
 import base64
 import datetime
 import hashlib
 import logging
 import secrets
+import struct
 import time
 from pprint import pformat as pf
-from typing import Any, Dict, Optional, Tuple, cast
+from typing import Any, cast
 
-from cryptography.hazmat.primitives import hashes, padding
+from cryptography.hazmat.primitives import padding
 from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
+from yarl import URL
 
 from .credentials import Credentials
 from .deviceconfig import DeviceConfig
-from .exceptions import AuthenticationException, SmartDeviceException
+from .exceptions import AuthenticationError, KasaException, _RetryableError
 from .httpclient import HttpClient
 from .json import loads as json_loads
-from .protocol import BaseTransport, md5
+from .protocol import DEFAULT_CREDENTIALS, BaseTransport, get_default_credentials, md5
 
 _LOGGER = logging.getLogger(__name__)
 
 
+ONE_DAY_SECONDS = 86400
+SESSION_EXPIRE_BUFFER_SECONDS = 60 * 20
+
+PACK_SIGNED_LONG = struct.Struct(">l").pack
+
+
 def _sha256(payload: bytes) -> bytes:
-    digest = hashes.Hash(hashes.SHA256())  # noqa: S303
-    digest.update(payload)
-    hash = digest.finalize()
-    return hash
+    return hashlib.sha256(payload).digest()  # noqa: S324
 
 
 def _sha1(payload: bytes) -> bytes:
-    digest = hashes.Hash(hashes.SHA1())  # noqa: S303
-    digest.update(payload)
-    return digest.finalize()
+    return hashlib.sha1(payload).digest()  # noqa: S324
 
 
 class KlapTransport(BaseTransport):
@@ -85,10 +90,8 @@ class KlapTransport(BaseTransport):
 
     DEFAULT_PORT: int = 80
     DISCOVERY_QUERY = {"system": {"get_sysinfo": None}}
-
-    KASA_SETUP_EMAIL = "kasa@tp-link.net"
-    KASA_SETUP_PASSWORD = "kasaSetup"  # noqa: S105
     SESSION_COOKIE_NAME = "TP_SESSIONID"
+    TIMEOUT_COOKIE_NAME = "TIMEOUT"
 
     def __init__(
         self,
@@ -98,7 +101,7 @@ class KlapTransport(BaseTransport):
         super().__init__(config=config)
 
         self._http_client = HttpClient(config)
-        self._local_seed: Optional[bytes] = None
+        self._local_seed: bytes | None = None
         if (
             not self._credentials or self._credentials.username is None
         ) and not self._credentials_hash:
@@ -108,18 +111,20 @@ class KlapTransport(BaseTransport):
             self._local_auth_owner = self.generate_owner_hash(self._credentials).hex()
         else:
             self._local_auth_hash = base64.b64decode(self._credentials_hash.encode())  # type: ignore[union-attr]
-        self._kasa_setup_auth_hash = None
+        self._default_credentials_auth_hash: dict[str, bytes] = {}
         self._blank_auth_hash = None
         self._handshake_lock = asyncio.Lock()
         self._query_lock = asyncio.Lock()
         self._handshake_done = False
 
-        self._encryption_session: Optional[KlapEncryptionSession] = None
-        self._session_expire_at: Optional[float] = None
+        self._encryption_session: KlapEncryptionSession | None = None
+        self._session_expire_at: float | None = None
 
-        self._session_cookie: Optional[Dict[str, Any]] = None
+        self._session_cookie: dict[str, Any] | None = None
 
         _LOGGER.debug("Created KLAP transport for %s", self._host)
+        self._app_url = URL(f"http://{self._host}:{self._port}/app")
+        self._request_url = self._app_url / "request"
 
     @property
     def default_port(self):
@@ -131,7 +136,7 @@ class KlapTransport(BaseTransport):
         """The hashed credentials used by the transport."""
         return base64.b64encode(self._local_auth_hash).decode()
 
-    async def perform_handshake1(self) -> Tuple[bytes, bytes, bytes]:
+    async def perform_handshake1(self) -> tuple[bytes, bytes, bytes]:
         """Perform handshake1."""
         local_seed: bytes = secrets.token_bytes(16)
 
@@ -141,7 +146,7 @@ class KlapTransport(BaseTransport):
 
         payload = local_seed
 
-        url = f"http://{self._host}/app/handshake1"
+        url = self._app_url / "handshake1"
 
         response_status, response_data = await self._http_client.post(url, data=payload)
 
@@ -156,7 +161,7 @@ class KlapTransport(BaseTransport):
             )
 
         if response_status != 200:
-            raise AuthenticationException(
+            raise KasaException(
                 f"Device {self._host} responded with {response_status} to handshake1"
             )
 
@@ -164,6 +169,12 @@ class KlapTransport(BaseTransport):
         remote_seed: bytes = response_data[0:16]
         server_hash = response_data[16:]
 
+        if len(server_hash) != 32:
+            raise KasaException(
+                f"Device {self._host} responded with unexpected klap response "
+                + f"{response_data!r} to handshake1"
+            )
+
         if _LOGGER.isEnabledFor(logging.DEBUG):
             _LOGGER.debug(
                 "Handshake1 success at %s. Host is %s, "
@@ -183,27 +194,27 @@ class KlapTransport(BaseTransport):
             _LOGGER.debug("handshake1 hashes match with expected credentials")
             return local_seed, remote_seed, self._local_auth_hash  # type: ignore
 
-        # Now check against the default kasa setup credentials
-        if not self._kasa_setup_auth_hash:
-            kasa_setup_creds = Credentials(
-                username=self.KASA_SETUP_EMAIL,
-                password=self.KASA_SETUP_PASSWORD,
-            )
-            self._kasa_setup_auth_hash = self.generate_auth_hash(kasa_setup_creds)
+        # Now check against the default setup credentials
+        for key, value in DEFAULT_CREDENTIALS.items():
+            if key not in self._default_credentials_auth_hash:
+                default_credentials = get_default_credentials(value)
+                self._default_credentials_auth_hash[key] = self.generate_auth_hash(
+                    default_credentials
+                )
 
-        kasa_setup_seed_auth_hash = self.handshake1_seed_auth_hash(
-            local_seed,
-            remote_seed,
-            self._kasa_setup_auth_hash,  # type: ignore
-        )
-
-        if kasa_setup_seed_auth_hash == server_hash:
-            _LOGGER.debug(
-                "Server response doesn't match our expected hash on ip %s"
-                + " but an authentication with kasa setup credentials matched",
-                self._host,
+            default_credentials_seed_auth_hash = self.handshake1_seed_auth_hash(
+                local_seed,
+                remote_seed,
+                self._default_credentials_auth_hash[key],  # type: ignore
             )
-            return local_seed, remote_seed, self._kasa_setup_auth_hash  # type: ignore
+
+            if default_credentials_seed_auth_hash == server_hash:
+                _LOGGER.debug(
+                    "Server response doesn't match our expected hash on ip %s"
+                    + f" but an authentication with {key} default credentials matched",
+                    self._host,
+                )
+                return local_seed, remote_seed, self._default_credentials_auth_hash[key]  # type: ignore
 
         # Finally check against blank credentials if not already blank
         blank_creds = Credentials()
@@ -227,16 +238,16 @@ class KlapTransport(BaseTransport):
 
         msg = f"Server response doesn't match our challenge on ip {self._host}"
         _LOGGER.debug(msg)
-        raise AuthenticationException(msg)
+        raise AuthenticationError(msg)
 
     async def perform_handshake2(
         self, local_seed, remote_seed, auth_hash
-    ) -> "KlapEncryptionSession":
+    ) -> KlapEncryptionSession:
         """Perform handshake2."""
         # Handshake 2 has the following payload:
         #    sha256(serverBytes | authenticator)
 
-        url = f"http://{self._host}/app/handshake2"
+        url = self._app_url / "handshake2"
 
         payload = self.handshake2_seed_auth_hash(local_seed, remote_seed, auth_hash)
 
@@ -257,7 +268,9 @@ class KlapTransport(BaseTransport):
             )
 
         if response_status != 200:
-            raise AuthenticationException(
+            # This shouldn't be caused by incorrect
+            # credentials so don't raise AuthenticationError
+            raise KasaException(
                 f"Device {self._host} responded with {response_status} to handshake2"
             )
 
@@ -274,14 +287,18 @@ class KlapTransport(BaseTransport):
         self._session_cookie = None
 
         local_seed, remote_seed, auth_hash = await self.perform_handshake1()
-        if cookie := self._http_client.get_cookie(  # type: ignore
-            self.SESSION_COOKIE_NAME
-        ):
+        http_client = self._http_client
+        if cookie := http_client.get_cookie(self.SESSION_COOKIE_NAME):  # type: ignore
             self._session_cookie = {self.SESSION_COOKIE_NAME: cookie}
         # The device returns a TIMEOUT cookie on handshake1 which
         # it doesn't like to get back so we store the one we want
-
-        self._session_expire_at = time.time() + 86400
+        timeout = int(
+            http_client.get_cookie(self.TIMEOUT_COOKIE_NAME) or ONE_DAY_SECONDS
+        )
+        # There is a 24 hour timeout on the session cookie
+        # but the clock on the device is not always accurate
+        # so we set the expiry to 24 hours from now minus a buffer
+        self._session_expire_at = time.time() + timeout - SESSION_EXPIRE_BUFFER_SECONDS
         self._encryption_session = await self.perform_handshake2(
             local_seed, remote_seed, auth_hash
         )
@@ -305,10 +322,8 @@ class KlapTransport(BaseTransport):
         if self._encryption_session is not None:
             payload, seq = self._encryption_session.encrypt(request.encode())
 
-        url = f"http://{self._host}/app/request"
-
         response_status, response_data = await self._http_client.post(
-            url,
+            self._request_url,
             params={"seq": seq},
             data=payload,
             cookies_dict=self._session_cookie,
@@ -320,16 +335,16 @@ class KlapTransport(BaseTransport):
             + f"Response status is {response_status}, Request was {request}"
         )
         if response_status != 200:
-            _LOGGER.error("Query failed after succesful authentication " + msg)
+            _LOGGER.error("Query failed after successful authentication " + msg)
             # If we failed with a security error, force a new handshake next time.
             if response_status == 403:
                 self._handshake_done = False
-                raise AuthenticationException(
+                raise _RetryableError(
                     f"Got a security error from {self._host} after handshake "
                     + "completed"
                 )
             else:
-                raise SmartDeviceException(
+                raise KasaException(
                     f"Device {self._host} responded with {response_status} to"
                     + f"request with seq {seq}"
                 )
@@ -351,10 +366,14 @@ class KlapTransport(BaseTransport):
             return json_payload
 
     async def close(self) -> None:
-        """Close the transport."""
-        self._handshake_done = False
+        """Close the http client and reset internal state."""
+        await self.reset()
         await self._http_client.close()
 
+    async def reset(self) -> None:
+        """Reset internal handshake state."""
+        self._handshake_done = False
+
     @staticmethod
     def generate_auth_hash(creds: Credentials):
         """Generate an md5 auth hash for the protocol on the supplied credentials."""
@@ -416,12 +435,15 @@ class KlapEncryptionSession:
     i.e. sequence number which the device expects to increment.
     """
 
+    _cipher: Cipher
+
     def __init__(self, local_seed, remote_seed, user_hash):
         self.local_seed = local_seed
         self.remote_seed = remote_seed
         self.user_hash = user_hash
         self._key = self._key_derive(local_seed, remote_seed, user_hash)
         (self._iv, self._seq) = self._iv_derive(local_seed, remote_seed, user_hash)
+        self._aes = algorithms.AES(self._key)
         self._sig = self._sig_derive(local_seed, remote_seed, user_hash)
 
     def _key_derive(self, local_seed, remote_seed, user_hash):
@@ -441,35 +463,31 @@ class KlapEncryptionSession:
         payload = b"ldk" + local_seed + remote_seed + user_hash
         return hashlib.sha256(payload).digest()[:28]
 
-    def _iv_seq(self):
-        seq = self._seq.to_bytes(4, "big", signed=True)
-        iv = self._iv + seq
-        return iv
+    def _generate_cipher(self):
+        iv_seq = self._iv + PACK_SIGNED_LONG(self._seq)
+        cbc = modes.CBC(iv_seq)
+        self._cipher = Cipher(self._aes, cbc)
 
     def encrypt(self, msg):
         """Encrypt the data and increment the sequence number."""
-        self._seq = self._seq + 1
+        self._seq += 1
+        self._generate_cipher()
+
         if isinstance(msg, str):
             msg = msg.encode("utf-8")
 
-        cipher = Cipher(algorithms.AES(self._key), modes.CBC(self._iv_seq()))
-        encryptor = cipher.encryptor()
+        encryptor = self._cipher.encryptor()
         padder = padding.PKCS7(128).padder()
         padded_data = padder.update(msg) + padder.finalize()
         ciphertext = encryptor.update(padded_data) + encryptor.finalize()
-
-        digest = hashes.Hash(hashes.SHA256())
-        digest.update(
-            self._sig + self._seq.to_bytes(4, "big", signed=True) + ciphertext
-        )
-        signature = digest.finalize()
-
+        signature = hashlib.sha256(
+            self._sig + PACK_SIGNED_LONG(self._seq) + ciphertext
+        ).digest()
         return (signature + ciphertext, self._seq)
 
     def decrypt(self, msg):
         """Decrypt the data."""
-        cipher = Cipher(algorithms.AES(self._key), modes.CBC(self._iv_seq()))
-        decryptor = cipher.decryptor()
+        decryptor = self._cipher.decryptor()
         dp = decryptor.update(msg[32:]) + decryptor.finalize()
         unpadder = padding.PKCS7(128).unpadder()
         plaintextbytes = unpadder.update(dp) + unpadder.finalize()
