@@ -4,24 +4,26 @@ Based on the work of https://github.com/petretiandrea/plugp100
 under compatible GNU GPL3 license.
 """
 
+from __future__ import annotations
+
 import asyncio
 import base64
 import logging
 import time
 import uuid
 from pprint import pformat as pf
-from typing import Any, Dict, Union
+from typing import Any
 
 from .exceptions import (
     SMART_AUTHENTICATION_ERRORS,
     SMART_RETRYABLE_ERRORS,
-    SMART_TIMEOUT_ERRORS,
-    AuthenticationException,
-    ConnectionException,
-    RetryableException,
-    SmartDeviceException,
+    AuthenticationError,
+    DeviceError,
+    KasaException,
     SmartErrorCode,
-    TimeoutException,
+    TimeoutError,
+    _ConnectionError,
+    _RetryableError,
 )
 from .json import dumps as json_dumps
 from .protocol import BaseProtocol, BaseTransport, md5
@@ -45,6 +47,9 @@ class SmartProtocol(BaseProtocol):
         self._terminal_uuid: str = base64.b64encode(md5(uuid.uuid4().bytes)).decode()
         self._request_id_generator = SnowflakeId(1, 1)
         self._query_lock = asyncio.Lock()
+        self._multi_request_batch_size = (
+            self._transport._config.batch_size or self.DEFAULT_MULTI_REQUEST_BATCH_SIZE
+        )
 
     def get_smart_request(self, method, params=None) -> str:
         """Get a request message as a string."""
@@ -57,40 +62,43 @@ class SmartProtocol(BaseProtocol):
         }
         return json_dumps(request)
 
-    async def query(self, request: Union[str, Dict], retry_count: int = 3) -> Dict:
+    async def query(self, request: str | dict, retry_count: int = 3) -> dict:
         """Query the device retrying for retry_count on failure."""
         async with self._query_lock:
             return await self._query(request, retry_count)
 
-    async def _query(self, request: Union[str, Dict], retry_count: int = 3) -> Dict:
+    async def _query(self, request: str | dict, retry_count: int = 3) -> dict:
         for retry in range(retry_count + 1):
             try:
-                return await self._execute_query(request, retry)
-            except ConnectionException as sdex:
+                return await self._execute_query(
+                    request, retry_count=retry, iterate_list_pages=True
+                )
+            except _ConnectionError as sdex:
                 if retry >= retry_count:
                     _LOGGER.debug("Giving up on %s after %s retries", self._host, retry)
                     raise sdex
                 continue
-            except AuthenticationException as auex:
+            except AuthenticationError as auex:
                 await self._transport.reset()
                 _LOGGER.debug(
                     "Unable to authenticate with %s, not retrying", self._host
                 )
                 raise auex
-            except RetryableException as ex:
-                await self._transport.reset()
-                if retry >= retry_count:
-                    _LOGGER.debug("Giving up on %s after %s retries", self._host, retry)
-                    raise ex
-                continue
-            except TimeoutException as ex:
+            except _RetryableError as ex:
                 await self._transport.reset()
                 if retry >= retry_count:
                     _LOGGER.debug("Giving up on %s after %s retries", self._host, retry)
                     raise ex
                 await asyncio.sleep(self.BACKOFF_SECONDS_AFTER_TIMEOUT)
                 continue
-            except SmartDeviceException as ex:
+            except TimeoutError as ex:
+                await self._transport.reset()
+                if retry >= retry_count:
+                    _LOGGER.debug("Giving up on %s after %s retries", self._host, retry)
+                    raise ex
+                await asyncio.sleep(self.BACKOFF_SECONDS_AFTER_TIMEOUT)
+                continue
+            except KasaException as ex:
                 await self._transport.reset()
                 _LOGGER.debug(
                     "Unable to query the device: %s, not retrying: %s",
@@ -100,23 +108,30 @@ class SmartProtocol(BaseProtocol):
                 raise ex
 
         # make mypy happy, this should never be reached..
-        raise SmartDeviceException("Query reached somehow to unreachable")
+        raise KasaException("Query reached somehow to unreachable")
 
-    async def _execute_multiple_query(self, request: Dict, retry_count: int) -> Dict:
+    async def _execute_multiple_query(self, requests: dict, retry_count: int) -> dict:
         debug_enabled = _LOGGER.isEnabledFor(logging.DEBUG)
-        multi_result: Dict[str, Any] = {}
+        multi_result: dict[str, Any] = {}
         smart_method = "multipleRequest"
-        requests = [
-            {"method": method, "params": params} for method, params in request.items()
+        multi_requests = [
+            {"method": method, "params": params} for method, params in requests.items()
         ]
 
-        end = len(requests)
+        end = len(multi_requests)
         # Break the requests down as there can be a size limit
-        step = (
-            self._transport._config.batch_size or self.DEFAULT_MULTI_REQUEST_BATCH_SIZE
-        )
+        step = self._multi_request_batch_size
+        if step == 1:
+            # If step is 1 do not send request batches
+            for request in multi_requests:
+                method = request["method"]
+                req = self.get_smart_request(method, request["params"])
+                resp = await self._transport.send(req)
+                self._handle_response_error_code(resp, method, raise_on_error=False)
+                multi_result[method] = resp["result"]
+            return multi_result
         for i in range(0, end, step):
-            requests_step = requests[i : i + step]
+            requests_step = multi_requests[i : i + step]
 
             smart_params = {"requests": requests_step}
             smart_request = self.get_smart_request(smart_method, smart_params)
@@ -128,22 +143,55 @@ class SmartProtocol(BaseProtocol):
                     pf(smart_request),
                 )
             response_step = await self._transport.send(smart_request)
+            batch_name = f"multi-request-batch-{i+1}"
             if debug_enabled:
                 _LOGGER.debug(
-                    "%s multi-request-batch-%s << %s",
+                    "%s %s << %s",
                     self._host,
-                    i + 1,
+                    batch_name,
                     pf(response_step),
                 )
-            self._handle_response_error_code(response_step)
+            try:
+                self._handle_response_error_code(response_step, batch_name)
+            except DeviceError as ex:
+                # P100 sometimes raises JSON_DECODE_FAIL_ERROR or INTERNAL_UNKNOWN_ERROR
+                # on batched request so disable batching
+                if (
+                    ex.error_code
+                    in {
+                        SmartErrorCode.JSON_DECODE_FAIL_ERROR,
+                        SmartErrorCode.INTERNAL_UNKNOWN_ERROR,
+                    }
+                    and self._multi_request_batch_size != 1
+                ):
+                    self._multi_request_batch_size = 1
+                    raise _RetryableError(
+                        "JSON Decode failure, multi requests disabled"
+                    ) from ex
+                raise ex
+
             responses = response_step["result"]["responses"]
             for response in responses:
-                self._handle_response_error_code(response)
+                method = response["method"]
+                self._handle_response_error_code(response, method, raise_on_error=False)
                 result = response.get("result", None)
-                multi_result[response["method"]] = result
+                await self._handle_response_lists(
+                    result, method, retry_count=retry_count
+                )
+                multi_result[method] = result
+        # Multi requests don't continue after errors so requery any missing
+        for method, params in requests.items():
+            if method not in multi_result:
+                resp = await self._transport.send(
+                    self.get_smart_request(method, params)
+                )
+                self._handle_response_error_code(resp, method, raise_on_error=False)
+                multi_result[method] = resp.get("result")
         return multi_result
 
-    async def _execute_query(self, request: Union[str, Dict], retry_count: int) -> Dict:
+    async def _execute_query(
+        self, request: str | dict, *, retry_count: int, iterate_list_pages: bool = True
+    ) -> dict:
         debug_enabled = _LOGGER.isEnabledFor(logging.DEBUG)
 
         if isinstance(request, dict):
@@ -172,29 +220,77 @@ class SmartProtocol(BaseProtocol):
                 pf(response_data),
             )
 
-        self._handle_response_error_code(response_data)
+        self._handle_response_error_code(response_data, smart_method)
 
         # Single set_ requests do not return a result
         result = response_data.get("result")
+        if iterate_list_pages and result:
+            await self._handle_response_lists(
+                result, smart_method, retry_count=retry_count
+            )
         return {smart_method: result}
 
-    def _handle_response_error_code(self, resp_dict: dict):
-        error_code = SmartErrorCode(resp_dict.get("error_code"))  # type: ignore[arg-type]
-        if error_code == SmartErrorCode.SUCCESS:
+    async def _handle_response_lists(
+        self, response_result: dict[str, Any], method, retry_count
+    ):
+        if (
+            response_result is None
+            or isinstance(response_result, SmartErrorCode)
+            or "start_index" not in response_result
+            or (list_sum := response_result.get("sum")) is None
+        ):
             return
+
+        response_list_name = next(
+            iter(
+                [
+                    key
+                    for key in response_result
+                    if isinstance(response_result[key], list)
+                ]
+            )
+        )
+        while (list_length := len(response_result[response_list_name])) < list_sum:
+            response = await self._execute_query(
+                {method: {"start_index": list_length}},
+                retry_count=retry_count,
+                iterate_list_pages=False,
+            )
+            next_batch = response[method]
+            # In case the device returns empty lists avoid infinite looping
+            if not next_batch[response_list_name]:
+                _LOGGER.error(
+                    f"Device {self._host} returned empty "
+                    + f"results list for method {method}"
+                )
+                break
+            response_result[response_list_name].extend(next_batch[response_list_name])
+
+    def _handle_response_error_code(self, resp_dict: dict, method, raise_on_error=True):
+        error_code_raw = resp_dict.get("error_code")
+        try:
+            error_code = SmartErrorCode.from_int(error_code_raw)
+        except ValueError:
+            _LOGGER.warning("Received unknown error code: %s", error_code_raw)
+            error_code = SmartErrorCode.INTERNAL_UNKNOWN_ERROR
+
+        if error_code is SmartErrorCode.SUCCESS:
+            return
+
+        if not raise_on_error:
+            resp_dict["result"] = error_code
+            return
+
         msg = (
             f"Error querying device: {self._host}: "
             + f"{error_code.name}({error_code.value})"
+            + f" for method: {method}"
         )
-        if method := resp_dict.get("method"):
-            msg += f" for method: {method}"
-        if error_code in SMART_TIMEOUT_ERRORS:
-            raise TimeoutException(msg, error_code=error_code)
         if error_code in SMART_RETRYABLE_ERRORS:
-            raise RetryableException(msg, error_code=error_code)
+            raise _RetryableError(msg, error_code=error_code)
         if error_code in SMART_AUTHENTICATION_ERRORS:
-            raise AuthenticationException(msg, error_code=error_code)
-        raise SmartDeviceException(msg, error_code=error_code)
+            raise AuthenticationError(msg, error_code=error_code)
+        raise DeviceError(msg, error_code=error_code)
 
     async def close(self) -> None:
         """Close the underlying transport."""
@@ -319,7 +415,7 @@ class _ChildProtocolWrapper(SmartProtocol):
 
         return smart_method, smart_params
 
-    async def query(self, request: Union[str, Dict], retry_count: int = 3) -> Dict:
+    async def query(self, request: str | dict, retry_count: int = 3) -> dict:
         """Wrap request inside control_child envelope."""
         method, params = self._get_method_and_params_for_request(request)
         request_data = {
@@ -337,10 +433,18 @@ class _ChildProtocolWrapper(SmartProtocol):
         result = response.get("control_child")
         # Unwrap responseData for control_child
         if result and (response_data := result.get("responseData")):
-            self._handle_response_error_code(response_data)
             result = response_data.get("result")
+            if result and (multi_responses := result.get("responses")):
+                ret_val = {}
+                for multi_response in multi_responses:
+                    method = multi_response["method"]
+                    self._handle_response_error_code(
+                        multi_response, method, raise_on_error=False
+                    )
+                    ret_val[method] = multi_response.get("result")
+                return ret_val
 
-        # TODO: handle multipleRequest unwrapping
+            self._handle_response_error_code(response_data, "control_child")
 
         return {method: result}
 
